# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import platform
from ctypes import c_void_p, c_int32, c_char_p, c_size_t, c_byte, c_int, c_uint32, c_uint64, c_ulong, c_long, c_bool
from ctypes import CDLL, POINTER, CFUNCTYPE, Structure

from .._ffi import FFIEngineError


__all__ = [
    'Security',
    'version',
    'version_info',
]


version = platform.mac_ver()[0]
version_info = tuple(map(int,  platform.mac_ver()[0].split('.')))

if version_info < (10, 7):
    raise OSError('Only OS X 10.7 and newer are supported, not %s.%s' % (version_info[0], version_info[1]))

security_path = '/System/Library/Frameworks/Security.framework/Security'

Security = CDLL(security_path, use_errno=True)

Boolean = c_bool
CFIndex = c_long
CFData = c_void_p
CFString = c_void_p
CFArray = c_void_p
CFDictionary = c_void_p
CFError = c_void_p
CFType = c_void_p
CFTypeID = c_ulong

CFTypeRef = POINTER(CFType)
CFAllocatorRef = c_void_p

OSStatus = c_int32

CFDataRef = POINTER(CFData)
CFStringRef = POINTER(CFString)
CFArrayRef = POINTER(CFArray)
CFDictionaryRef = POINTER(CFDictionary)
CFErrorRef = POINTER(CFError)

SecKeyRef = POINTER(c_void_p)
SecCertificateRef = POINTER(c_void_p)
SecTransformRef = POINTER(c_void_p)
SecRandomRef = c_void_p
SecTrustSettingsDomain = c_uint32
SecItemImportExportFlags = c_uint32
SecKeyImportExportFlags = c_uint32
SecExternalFormat = c_uint32
SecExternalItemType = c_uint32
SecPadding = c_uint32
SSLProtocol = c_uint32
SSLCipherSuite = c_uint32
SecPolicyRef = POINTER(c_void_p)
CSSM_CC_HANDLE = c_uint64
CSSM_ALGORITHMS = c_uint32
CSSM_KEYUSE = c_uint32
SecAccessRef = POINTER(c_void_p)
SecKeychainRef = POINTER(c_void_p)
SSLContextRef = POINTER(c_void_p)
SecTrustRef = POINTER(c_void_p)
SSLConnectionRef = c_uint32
SecTrustResultType = c_uint32
SecTrustOptionFlags = c_uint32
SecPolicySearchRef = c_void_p
CSSM_CERT_TYPE = c_uint32


class CSSM_DATA(Structure):  # noqa
    _fields_ = [
        ('Length', c_uint32),
        ('Data', c_char_p)
    ]


CSSM_OID = CSSM_DATA


class CSSM_APPLE_TP_OCSP_OPTIONS(Structure):  # noqa
    _fields_ = [
        ('Version', c_uint32),
        ('Flags', c_uint32),
        ('LocalResponder', POINTER(CSSM_DATA)),
        ('LocalResponderCert', POINTER(CSSM_DATA)),
    ]


class CSSM_APPLE_TP_CRL_OPTIONS(Structure):  # noqa
    _fields_ = [
        ('Version', c_uint32),
        ('CrlFlags', c_uint32),
        ('crlStore', c_void_p),
    ]


class SecItemImportExportKeyParameters(Structure):
    _fields_ = [
        ('version', c_uint32),
        ('flags', SecKeyImportExportFlags),
        ('passphrase', CFTypeRef),
        ('alertTitle', CFStringRef),
        ('alertPrompt', CFStringRef),
        ('accessRef', SecAccessRef),
        ('keyUsage', CFArrayRef),
        ('keyAttributes', CFArrayRef),
    ]


try:
    Security.SecKeychainCreate.argtypes = [
        c_char_p,
        c_uint32,
        c_void_p,
        Boolean,
        SecAccessRef,
        POINTER(SecKeychainRef)
    ]
    Security.SecKeychainCreate.restype = OSStatus

    Security.SecKeychainDelete.argtypes = [SecKeychainRef]
    Security.SecKeychainDelete.restype = OSStatus

    Security.SecRandomCopyBytes.argtypes = [
        SecRandomRef,
        c_size_t,
        c_char_p
    ]
    Security.SecRandomCopyBytes.restype = c_int

    Security.SecKeyCreateFromData.argtypes = [
        CFDictionaryRef,
        CFDataRef,
        POINTER(CFErrorRef)
    ]
    Security.SecKeyCreateFromData.restype = SecKeyRef

    Security.SecEncryptTransformCreate.argtypes = [
        SecKeyRef,
        POINTER(CFErrorRef)
    ]
    Security.SecEncryptTransformCreate.restype = SecTransformRef

    Security.SecDecryptTransformCreate.argtypes = [
        SecKeyRef,
        POINTER(CFErrorRef)
    ]
    Security.SecDecryptTransformCreate.restype = SecTransformRef

    Security.SecTransformSetAttribute.argtypes = [
        SecTransformRef,
        CFStringRef,
        CFTypeRef,
        POINTER(CFErrorRef)
    ]
    Security.SecTransformSetAttribute.restype = Boolean

    Security.SecTransformExecute.argtypes = [
        SecTransformRef,
        POINTER(CFErrorRef)
    ]
    Security.SecTransformExecute.restype = CFTypeRef

    Security.SecVerifyTransformCreate.argtypes = [
        SecKeyRef,
        CFDataRef,
        POINTER(CFErrorRef)
    ]
    Security.SecVerifyTransformCreate.restype = SecTransformRef

    Security.SecSignTransformCreate.argtypes = [
        SecKeyRef,
        POINTER(CFErrorRef)
    ]
    Security.SecSignTransformCreate.restype = SecTransformRef

    Security.SecCertificateCreateWithData.argtypes = [
        CFAllocatorRef,
        CFDataRef
    ]
    Security.SecCertificateCreateWithData.restype = SecCertificateRef

    Security.SecCertificateCopyPublicKey.argtypes = [
        SecCertificateRef,
        POINTER(SecKeyRef)
    ]
    Security.SecCertificateCopyPublicKey.restype = OSStatus

    Security.SecCopyErrorMessageString.argtypes = [
        OSStatus,
        c_void_p
    ]
    Security.SecCopyErrorMessageString.restype = CFStringRef

    Security.SecTrustCopyAnchorCertificates.argtypes = [
        POINTER(CFArrayRef)
    ]
    Security.SecTrustCopyAnchorCertificates.restype = OSStatus

    Security.SecCertificateCopyData.argtypes = [
        SecCertificateRef
    ]
    Security.SecCertificateCopyData.restype = CFDataRef

    Security.SecTrustSettingsCopyCertificates.argtypes = [
        SecTrustSettingsDomain,
        POINTER(CFArrayRef)
    ]
    Security.SecTrustSettingsCopyCertificates.restype = OSStatus

    Security.SecTrustSettingsCopyTrustSettings.argtypes = [
        SecCertificateRef,
        SecTrustSettingsDomain,
        POINTER(CFArrayRef)
    ]
    Security.SecTrustSettingsCopyTrustSettings.restype = OSStatus

    Security.SecPolicyCopyProperties.argtypes = [
        SecPolicyRef
    ]
    Security.SecPolicyCopyProperties.restype = CFDictionaryRef

    Security.SecPolicyGetTypeID.argtypes = []
    Security.SecPolicyGetTypeID.restype = CFTypeID

    Security.SecKeyEncrypt.argtypes = [
        SecKeyRef,
        SecPadding,
        c_char_p,
        c_size_t,
        c_char_p,
        POINTER(c_size_t)
    ]
    Security.SecKeyEncrypt.restype = OSStatus

    Security.SecKeyDecrypt.argtypes = [
        SecKeyRef,
        SecPadding,
        c_char_p,
        c_size_t,
        c_char_p,
        POINTER(c_size_t)
    ]
    Security.SecKeyDecrypt.restype = OSStatus

    Security.SecKeyRawSign.argtypes = [
        SecKeyRef,
        SecPadding,
        c_char_p,
        c_size_t,
        c_char_p,
        POINTER(c_size_t)
    ]
    Security.SecKeyRawSign.restype = OSStatus

    Security.SecKeyRawVerify.argtypes = [
        SecKeyRef,
        SecPadding,
        c_char_p,
        c_size_t,
        c_char_p,
        c_size_t
    ]
    Security.SecKeyRawVerify.restype = OSStatus

    Security.SecAccessCreate.argtypes = [
        CFStringRef,
        CFArrayRef,
        POINTER(SecAccessRef)
    ]
    Security.SecAccessCreate.restype = OSStatus

    Security.SecKeyCreatePair.argtypes = [
        SecKeychainRef,
        CSSM_ALGORITHMS,
        c_uint32,
        CSSM_CC_HANDLE,
        CSSM_KEYUSE,
        c_uint32,
        CSSM_KEYUSE,
        c_uint32,
        SecAccessRef,
        POINTER(SecKeyRef),
        POINTER(SecKeyRef)
    ]
    Security.SecKeyCreatePair.restype = OSStatus

    Security.SecItemImport.argtypes = [
        CFDataRef,
        CFStringRef,
        POINTER(SecExternalFormat),
        POINTER(SecExternalItemType),
        SecItemImportExportFlags,
        POINTER(SecItemImportExportKeyParameters),
        SecKeychainRef,
        POINTER(CFArrayRef)
    ]
    Security.SecItemImport.restype = OSStatus

    Security.SecItemExport.argtypes = [
        CFTypeRef,
        SecExternalFormat,
        SecItemImportExportFlags,
        POINTER(SecItemImportExportKeyParameters),
        POINTER(CFDataRef)
    ]
    Security.SecItemExport.restype = OSStatus

    Security.SecKeychainItemDelete.argtypes = [
        SecKeyRef
    ]
    Security.SecKeychainItemDelete.restype = OSStatus

    SSLReadFunc = CFUNCTYPE(OSStatus, SSLConnectionRef, POINTER(c_byte), POINTER(c_size_t))
    SSLWriteFunc = CFUNCTYPE(OSStatus, SSLConnectionRef, POINTER(c_byte), POINTER(c_size_t))

    Security.SSLSetIOFuncs.argtypes = [
        SSLContextRef,
        SSLReadFunc,
        SSLWriteFunc
    ]
    Security.SSLSetIOFuncs.restype = OSStatus

    Security.SSLSetPeerID.argtypes = [
        SSLContextRef,
        c_char_p,
        c_size_t
    ]
    Security.SSLSetPeerID.restype = OSStatus

    Security.SSLSetCertificateAuthorities.argtypes = [
        SSLContextRef,
        CFTypeRef,
        Boolean
    ]
    Security.SSLSetCertificateAuthorities.restype = OSStatus

    Security.SecTrustSetPolicies.argtypes = [
        SecTrustRef,
        CFArrayRef
    ]
    Security.SecTrustSetPolicies.restype = OSStatus

    Security.SecPolicyCreateSSL.argtypes = [
        Boolean,
        CFStringRef
    ]
    Security.SecPolicyCreateSSL.restype = SecPolicyRef

    Security.SecPolicySearchCreate.argtypes = [
        CSSM_CERT_TYPE,
        POINTER(CSSM_OID),
        POINTER(CSSM_DATA),
        POINTER(SecPolicySearchRef)
    ]
    Security.SecPolicySearchCreate.restype = OSStatus

    Security.SecPolicySearchCopyNext.argtypes = [
        SecPolicySearchRef,
        POINTER(SecPolicyRef)
    ]
    Security.SecPolicySearchCopyNext.restype = OSStatus

    Security.SecPolicySetValue.argtypes = [
        SecPolicyRef,
        POINTER(CSSM_DATA)
    ]
    Security.SecPolicySetValue.restype = OSStatus

    Security.SSLSetConnection.argtypes = [
        SSLContextRef,
        SSLConnectionRef
    ]
    Security.SSLSetConnection.restype = OSStatus

    Security.SSLSetPeerDomainName.argtypes = [
        SSLContextRef,
        c_char_p,
        c_size_t
    ]
    Security.SSLSetPeerDomainName.restype = OSStatus

    Security.SSLHandshake.argtypes = [
        SSLContextRef
    ]
    Security.SSLHandshake.restype = OSStatus

    Security.SSLGetBufferedReadSize.argtypes = [
        SSLContextRef,
        POINTER(c_size_t)
    ]
    Security.SSLGetBufferedReadSize.restype = OSStatus

    Security.SSLRead.argtypes = [
        SSLContextRef,
        c_char_p,
        c_size_t,
        POINTER(c_size_t)
    ]
    Security.SSLRead.restype = OSStatus

    Security.SSLWrite.argtypes = [
        SSLContextRef,
        c_char_p,
        c_size_t,
        POINTER(c_size_t)
    ]
    Security.SSLWrite.restype = OSStatus

    Security.SSLClose.argtypes = [
        SSLContextRef
    ]
    Security.SSLClose.restype = OSStatus

    Security.SSLGetNumberSupportedCiphers.argtypes = [
        SSLContextRef,
        POINTER(c_size_t)
    ]
    Security.SSLGetNumberSupportedCiphers.restype = OSStatus

    Security.SSLGetSupportedCiphers.argtypes = [
        SSLContextRef,
        POINTER(SSLCipherSuite),
        POINTER(c_size_t)
    ]
    Security.SSLGetSupportedCiphers.restype = OSStatus

    Security.SSLSetEnabledCiphers.argtypes = [
        SSLContextRef,
        POINTER(SSLCipherSuite),
        c_size_t
    ]
    Security.SSLSetEnabledCiphers.restype = OSStatus

    Security.SSLGetNumberEnabledCiphers.argtype = [
        SSLContextRef,
        POINTER(c_size_t)
    ]
    Security.SSLGetNumberEnabledCiphers.restype = OSStatus

    Security.SSLGetEnabledCiphers.argtypes = [
        SSLContextRef,
        POINTER(SSLCipherSuite),
        POINTER(c_size_t)
    ]
    Security.SSLGetEnabledCiphers.restype = OSStatus

    Security.SSLGetNegotiatedCipher.argtypes = [
        SSLContextRef,
        POINTER(SSLCipherSuite)
    ]
    Security.SSLGetNegotiatedCipher.restype = OSStatus

    Security.SSLGetNegotiatedProtocolVersion.argtypes = [
        SSLContextRef,
        POINTER(SSLProtocol)
    ]
    Security.SSLGetNegotiatedProtocolVersion.restype = OSStatus

    Security.SSLCopyPeerTrust.argtypes = [
        SSLContextRef,
        POINTER(SecTrustRef)
    ]
    Security.SSLCopyPeerTrust.restype = OSStatus

    Security.SecTrustGetCssmResultCode.argtypes = [
        SecTrustRef,
        POINTER(OSStatus)
    ]
    Security.SecTrustGetCssmResultCode.restype = OSStatus

    Security.SecTrustGetCertificateCount.argtypes = [
        SecTrustRef
    ]
    Security.SecTrustGetCertificateCount.restype = CFIndex

    Security.SecTrustGetCertificateAtIndex.argtypes = [
        SecTrustRef,
        CFIndex
    ]
    Security.SecTrustGetCertificateAtIndex.restype = SecCertificateRef

    Security.SecTrustSetAnchorCertificates.argtypes = [
        SecTrustRef,
        CFArrayRef
    ]
    Security.SecTrustSetAnchorCertificates.restype = OSStatus

    Security.SecTrustSetAnchorCertificatesOnly.argstypes = [
        SecTrustRef,
        Boolean
    ]
    Security.SecTrustSetAnchorCertificatesOnly.restype = OSStatus

    Security.SecTrustEvaluate.argtypes = [
        SecTrustRef,
        POINTER(SecTrustResultType)
    ]
    Security.SecTrustEvaluate.restype = OSStatus

    if version_info < (10, 8):
        Security.SSLNewContext.argtypes = [
            Boolean,
            POINTER(SSLContextRef)
        ]
        Security.SSLNewContext.restype = OSStatus

        Security.SSLDisposeContext.argtypes = [
            SSLContextRef
        ]
        Security.SSLDisposeContext.restype = OSStatus

        Security.SSLSetEnableCertVerify.argtypes = [
            SSLContextRef,
            Boolean
        ]
        Security.SSLSetEnableCertVerify.restype = OSStatus

        Security.SSLSetProtocolVersionEnabled.argtypes = [
            SSLContextRef,
            SSLProtocol,
            Boolean
        ]
        Security.SSLSetProtocolVersionEnabled.restype = OSStatus

    else:
        SSLProtocolSide = c_uint32
        SSLConnectionType = c_uint32
        SSLSessionOption = c_uint32

        Security.SSLCreateContext.argtypes = [
            CFAllocatorRef,
            SSLProtocolSide,
            SSLConnectionType
        ]
        Security.SSLCreateContext.restype = SSLContextRef

        Security.SSLSetSessionOption.argtypes = [
            SSLContextRef,
            SSLSessionOption,
            Boolean
        ]
        Security.SSLSetSessionOption.restype = OSStatus

        Security.SSLSetProtocolVersionMin.argtypes = [
            SSLContextRef,
            SSLProtocol
        ]
        Security.SSLSetProtocolVersionMin.restype = OSStatus

        Security.SSLSetProtocolVersionMax.argtypes = [
            SSLContextRef,
            SSLProtocol
        ]
        Security.SSLSetProtocolVersionMax.restype = OSStatus

    setattr(Security, 'SSLReadFunc', SSLReadFunc)
    setattr(Security, 'SSLWriteFunc', SSLWriteFunc)
    setattr(Security, 'SSLContextRef', SSLContextRef)
    setattr(Security, 'SSLProtocol', SSLProtocol)
    setattr(Security, 'SSLCipherSuite', SSLCipherSuite)
    setattr(Security, 'SecTrustRef', SecTrustRef)
    setattr(Security, 'SecTrustResultType', SecTrustResultType)
    setattr(Security, 'OSStatus', OSStatus)

    setattr(Security, 'SecAccessRef', SecAccessRef)
    setattr(Security, 'SecKeychainRef', SecKeychainRef)
    setattr(Security, 'SecKeyRef', SecKeyRef)

    setattr(Security, 'SecPolicySearchRef', SecPolicySearchRef)
    setattr(Security, 'SecPolicyRef', SecPolicyRef)

    setattr(Security, 'CSSM_DATA', CSSM_DATA)
    setattr(Security, 'CSSM_OID', CSSM_OID)
    setattr(Security, 'CSSM_APPLE_TP_OCSP_OPTIONS', CSSM_APPLE_TP_OCSP_OPTIONS)
    setattr(Security, 'CSSM_APPLE_TP_CRL_OPTIONS', CSSM_APPLE_TP_CRL_OPTIONS)
    setattr(Security, 'SecItemImportExportKeyParameters', SecItemImportExportKeyParameters)

    setattr(Security, 'kSecRandomDefault', SecRandomRef.in_dll(Security, 'kSecRandomDefault'))

    setattr(Security, 'kSecPaddingKey', CFStringRef.in_dll(Security, 'kSecPaddingKey'))
    setattr(Security, 'kSecPaddingPKCS7Key', CFStringRef.in_dll(Security, 'kSecPaddingPKCS7Key'))
    setattr(Security, 'kSecPaddingPKCS5Key', CFStringRef.in_dll(Security, 'kSecPaddingPKCS5Key'))
    setattr(Security, 'kSecPaddingPKCS1Key', CFStringRef.in_dll(Security, 'kSecPaddingPKCS1Key'))
    setattr(Security, 'kSecPaddingOAEPKey', CFStringRef.in_dll(Security, 'kSecPaddingOAEPKey'))
    setattr(Security, 'kSecPaddingNoneKey', CFStringRef.in_dll(Security, 'kSecPaddingNoneKey'))
    setattr(Security, 'kSecModeCBCKey', CFStringRef.in_dll(Security, 'kSecModeCBCKey'))
    setattr(
        Security,
        'kSecTransformInputAttributeName',
        CFStringRef.in_dll(Security, 'kSecTransformInputAttributeName')
    )
    setattr(Security, 'kSecDigestTypeAttribute', CFStringRef.in_dll(Security, 'kSecDigestTypeAttribute'))
    setattr(Security, 'kSecDigestLengthAttribute', CFStringRef.in_dll(Security, 'kSecDigestLengthAttribute'))
    setattr(Security, 'kSecIVKey', CFStringRef.in_dll(Security, 'kSecIVKey'))

    setattr(Security, 'kSecAttrIsExtractable', CFStringRef.in_dll(Security, 'kSecAttrIsExtractable'))

    setattr(Security, 'kSecDigestSHA1', CFStringRef.in_dll(Security, 'kSecDigestSHA1'))
    setattr(Security, 'kSecDigestSHA2', CFStringRef.in_dll(Security, 'kSecDigestSHA2'))
    setattr(Security, 'kSecDigestMD5', CFStringRef.in_dll(Security, 'kSecDigestMD5'))

    setattr(Security, 'kSecAttrKeyType', CFStringRef.in_dll(Security, 'kSecAttrKeyType'))

    setattr(Security, 'kSecAttrKeyTypeRSA', CFTypeRef.in_dll(Security, 'kSecAttrKeyTypeRSA'))
    setattr(Security, 'kSecAttrKeyTypeDSA', CFTypeRef.in_dll(Security, 'kSecAttrKeyTypeDSA'))
    setattr(Security, 'kSecAttrKeyTypeECDSA', CFTypeRef.in_dll(Security, 'kSecAttrKeyTypeECDSA'))

    setattr(Security, 'kSecAttrKeySizeInBits', CFStringRef.in_dll(Security, 'kSecAttrKeySizeInBits'))
    setattr(Security, 'kSecAttrLabel', CFStringRef.in_dll(Security, 'kSecAttrLabel'))

    setattr(Security, 'kSecAttrCanSign', CFTypeRef.in_dll(Security, 'kSecAttrCanSign'))
    setattr(Security, 'kSecAttrCanVerify', CFTypeRef.in_dll(Security, 'kSecAttrCanVerify'))

    setattr(Security, 'kSecAttrKeyTypeAES', CFTypeRef.in_dll(Security, 'kSecAttrKeyTypeAES'))
    setattr(Security, 'kSecAttrKeyTypeRC4', CFTypeRef.in_dll(Security, 'kSecAttrKeyTypeRC4'))
    setattr(Security, 'kSecAttrKeyTypeRC2', CFTypeRef.in_dll(Security, 'kSecAttrKeyTypeRC2'))
    setattr(Security, 'kSecAttrKeyType3DES', CFTypeRef.in_dll(Security, 'kSecAttrKeyType3DES'))
    setattr(Security, 'kSecAttrKeyTypeDES', CFTypeRef.in_dll(Security, 'kSecAttrKeyTypeDES'))

except (AttributeError):
    raise FFIEngineError('Error initializing ctypes')



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import sys
import hashlib
import hmac
import struct

from ._asn1 import int_from_bytes, int_to_bytes
from ._errors import pretty_message
from ._types import type_name, byte_cls, int_types

if sys.version_info < (3,):
    chr_cls = chr

else:
    def chr_cls(num):
        return bytes([num])


__all__ = [
    'pbkdf2',
]


def pbkdf2(hash_algorithm, password, salt, iterations, key_length):
    """
    Implements PBKDF2 from PKCS#5 v2.2 in pure Python

    :param hash_algorithm:
        The string name of the hash algorithm to use: "md5", "sha1", "sha224",
        "sha256", "sha384", "sha512"

    :param password:
        A byte string of the password to use an input to the KDF

    :param salt:
        A cryptographic random byte string

    :param iterations:
        The numbers of iterations to use when deriving the key

    :param key_length:
        The length of the desired key in bytes

    :return:
        The derived key as a byte string
    """

    if not isinstance(password, byte_cls):
        raise TypeError(pretty_message(
            '''
            password must be a byte string, not %s
            ''',
            type_name(password)
        ))

    if not isinstance(salt, byte_cls):
        raise TypeError(pretty_message(
            '''
            salt must be a byte string, not %s
            ''',
            type_name(salt)
        ))

    if not isinstance(iterations, int_types):
        raise TypeError(pretty_message(
            '''
            iterations must be an integer, not %s
            ''',
            type_name(iterations)
        ))

    if iterations < 1:
        raise ValueError(pretty_message(
            '''
            iterations must be greater than 0 - is %s
            ''',
            repr(iterations)
        ))

    if not isinstance(key_length, int_types):
        raise TypeError(pretty_message(
            '''
            key_length must be an integer, not %s
            ''',
            type_name(key_length)
        ))

    if key_length < 1:
        raise ValueError(pretty_message(
            '''
            key_length must be greater than 0 - is %s
            ''',
            repr(key_length)
        ))

    if hash_algorithm not in set(['md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512']):
        raise ValueError(pretty_message(
            '''
            hash_algorithm must be one of "md5", "sha1", "sha224", "sha256",
            "sha384", "sha512", not %s
            ''',
            repr(hash_algorithm)
        ))

    algo = getattr(hashlib, hash_algorithm)

    hash_length = {
        'md5': 16,
        'sha1': 20,
        'sha224': 28,
        'sha256': 32,
        'sha384': 48,
        'sha512': 64
    }[hash_algorithm]

    original_hmac = hmac.new(password, None, algo)

    block = 1
    output = b''

    while len(output) < key_length:
        prf = original_hmac.copy()
        prf.update(salt + struct.pack(b'>I', block))
        last = prf.digest()

        u = int_from_bytes(last)

        for _ in range(iterations-1):
            prf = original_hmac.copy()
            prf.update(last)
            last = prf.digest()
            u ^= int_from_bytes(last)

        output += int_to_bytes(u, width=hash_length)
        block += 1

    return output[0:key_length]


pbkdf2.pure_python = True



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import sys

from .._ffi import register_ffi
from .._types import str_cls
from ..errors import LibraryNotFoundError

import cffi


__all__ = [
    'get_error',
    'secur32',
]


ffi = cffi.FFI()
if cffi.__version_info__ >= (0, 9):
    ffi.set_unicode(True)
if sys.maxsize > 2 ** 32:
    ffi.cdef("typedef uint64_t ULONG_PTR;")
else:
    ffi.cdef("typedef unsigned long ULONG_PTR;")
ffi.cdef("""
    typedef HANDLE HCERTSTORE;
    typedef unsigned int ALG_ID;
    typedef WCHAR SEC_WCHAR;
    typedef unsigned long SECURITY_STATUS;
    typedef void *LUID;
    typedef void *SEC_GET_KEY_FN;

    typedef struct _SecHandle {
        ULONG_PTR dwLower;
        ULONG_PTR dwUpper;
    } SecHandle;
    typedef SecHandle CredHandle;
    typedef SecHandle CtxtHandle;

    typedef struct _SCHANNEL_CRED {
        DWORD dwVersion;
        DWORD cCreds;
        void *paCred;
        HCERTSTORE hRootStore;
        DWORD cMappers;
        void **aphMappers;
        DWORD cSupportedAlgs;
        ALG_ID *palgSupportedAlgs;
        DWORD grbitEnabledProtocols;
        DWORD dwMinimumCipherStrength;
        DWORD dwMaximumCipherStrength;
        DWORD dwSessionLifespan;
        DWORD dwFlags;
        DWORD dwCredFormat;
    } SCHANNEL_CRED;

    typedef struct _TimeStamp {
        DWORD dwLowDateTime;
        DWORD dwHighDateTime;
    } TimeStamp;

    typedef struct _SecBuffer {
        ULONG cbBuffer;
        ULONG BufferType;
        BYTE *pvBuffer;
    } SecBuffer;

    typedef struct _SecBufferDesc {
        ULONG ulVersion;
        ULONG cBuffers;
        SecBuffer *pBuffers;
    } SecBufferDesc;

    typedef struct _SecPkgContext_StreamSizes {
        ULONG cbHeader;
        ULONG cbTrailer;
        ULONG cbMaximumMessage;
        ULONG cBuffers;
        ULONG cbBlockSize;
    } SecPkgContext_StreamSizes;

    typedef struct _CERT_CONTEXT {
        DWORD dwCertEncodingType;
        BYTE *pbCertEncoded;
        DWORD cbCertEncoded;
        void *pCertInfo;
        HCERTSTORE hCertStore;
    } CERT_CONTEXT;

    typedef struct _SecPkgContext_ConnectionInfo {
        DWORD dwProtocol;
        ALG_ID aiCipher;
        DWORD dwCipherStrength;
        ALG_ID aiHash;
        DWORD dwHashStrength;
        ALG_ID aiExch;
        DWORD dwExchStrength;
    } SecPkgContext_ConnectionInfo;

    SECURITY_STATUS AcquireCredentialsHandleW(SEC_WCHAR *pszPrincipal, SEC_WCHAR *pszPackage, ULONG fCredentialUse,
                    LUID *pvLogonID, void *pAuthData, SEC_GET_KEY_FN pGetKeyFn, void *pvGetKeyArgument,
                    CredHandle *phCredential, TimeStamp *ptsExpiry);
    SECURITY_STATUS FreeCredentialsHandle(CredHandle *phCredential);
    SECURITY_STATUS InitializeSecurityContextW(CredHandle *phCredential, CtxtHandle *phContext,
                    SEC_WCHAR *pszTargetName, ULONG fContextReq, ULONG Reserved1, ULONG TargetDataRep,
                    SecBufferDesc *pInput, ULONG Reserved2, CtxtHandle *phNewContext, SecBufferDesc *pOutput,
                    ULONG *pfContextAttr, TimeStamp *ptsExpiry);
    SECURITY_STATUS FreeContextBuffer(void *pvContextBuffer);
    SECURITY_STATUS ApplyControlToken(CtxtHandle *phContext, SecBufferDesc *pInput);
    SECURITY_STATUS DeleteSecurityContext(CtxtHandle *phContext);
    SECURITY_STATUS QueryContextAttributesW(CtxtHandle *phContext, ULONG ulAttribute, void *pBuffer);
    SECURITY_STATUS EncryptMessage(CtxtHandle *phContext, ULONG fQOP, SecBufferDesc *pMessage, ULONG MessageSeqNo);
    SECURITY_STATUS DecryptMessage(CtxtHandle *phContext, SecBufferDesc *pMessage, ULONG MessageSeqNo, ULONG *pfQOP);
""")


try:
    secur32 = ffi.dlopen('secur32.dll')
    register_ffi(secur32, ffi)

except (OSError) as e:
    if str_cls(e).find('cannot load library') != -1:
        raise LibraryNotFoundError('secur32.dll could not be found')
    raise


def get_error():
    return ffi.getwinerror()



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

from .. import ffi
from .._ffi import null
from ..errors import TLSDisconnectError, TLSGracefulDisconnectError

if ffi() == 'cffi':
    from ._security_cffi import Security, version_info as osx_version_info
    from ._core_foundation_cffi import CoreFoundation, CFHelpers
else:
    from ._security_ctypes import Security, version_info as osx_version_info
    from ._core_foundation_ctypes import CoreFoundation, CFHelpers


__all__ = [
    'handle_sec_error',
    'osx_version_info',
    'Security',
    'SecurityConst',
]


def handle_sec_error(error, exception_class=None):
    """
    Checks a Security OSStatus error code and throws an exception if there is an
    error to report

    :param error:
        An OSStatus

    :param exception_class:
        The exception class to use for the exception if an error occurred

    :raises:
        OSError - when the OSStatus contains an error
    """

    if error == 0:
        return

    if error in set([SecurityConst.errSSLClosedNoNotify, SecurityConst.errSSLClosedAbort]):
        raise TLSDisconnectError('The remote end closed the connection')
    if error == SecurityConst.errSSLClosedGraceful:
        raise TLSGracefulDisconnectError('The remote end closed the connection')

    cf_error_string = Security.SecCopyErrorMessageString(error, null())
    output = CFHelpers.cf_string_to_unicode(cf_error_string)
    CoreFoundation.CFRelease(cf_error_string)

    if output is None or output == '':
        output = 'OSStatus %s' % error

    if exception_class is None:
        exception_class = OSError

    raise exception_class(output)


def _extract_policy_properties(value):
    properties_dict = Security.SecPolicyCopyProperties(value)
    return CFHelpers.cf_dictionary_to_dict(properties_dict)


CFHelpers.register_native_mapping(
    Security.SecPolicyGetTypeID(),
    _extract_policy_properties
)


class SecurityConst():
    kSecTrustSettingsDomainUser = 0
    kSecTrustSettingsDomainAdmin = 1
    kSecTrustSettingsDomainSystem = 2

    kSecTrustResultProceed = 1
    kSecTrustResultUnspecified = 4
    kSecTrustOptionImplicitAnchors = 0x00000040

    kSecFormatOpenSSL = 1

    kSecItemTypePrivateKey = 1
    kSecItemTypePublicKey = 2

    kSSLSessionOptionBreakOnServerAuth = 0

    kSSLProtocol2 = 1
    kSSLProtocol3 = 2
    kTLSProtocol1 = 4
    kTLSProtocol11 = 7
    kTLSProtocol12 = 8

    kSSLClientSide = 1
    kSSLStreamType = 0

    errSSLProtocol = -9800
    errSSLWouldBlock = -9803
    errSSLClosedGraceful = -9805
    errSSLClosedNoNotify = -9816
    errSSLClosedAbort = -9806

    errSSLXCertChainInvalid = -9807
    errSSLCrypto = -9809
    errSSLInternal = -9810
    errSSLCertExpired = -9814
    errSSLCertNotYetValid = -9815
    errSSLUnknownRootCert = -9812
    errSSLNoRootCert = -9813
    errSSLHostNameMismatch = -9843
    errSSLPeerHandshakeFail = -9824
    errSSLPeerProtocolVersion = -9836
    errSSLPeerUserCancelled = -9839
    errSSLWeakPeerEphemeralDHKey = -9850
    errSSLServerAuthCompleted = -9841
    errSSLRecordOverflow = -9847

    CSSMERR_APPLETP_HOSTNAME_MISMATCH = -2147408896
    CSSMERR_TP_CERT_EXPIRED = -2147409654
    CSSMERR_TP_CERT_NOT_VALID_YET = -2147409653
    CSSMERR_TP_CERT_REVOKED = -2147409652
    CSSMERR_TP_NOT_TRUSTED = -2147409622
    CSSMERR_TP_CERT_SUSPENDED = -2147409651

    CSSM_CERT_X_509v3 = 0x00000004

    APPLE_TP_REVOCATION_CRL = b'*\x86H\x86\xf7cd\x01\x06'
    APPLE_TP_REVOCATION_OCSP = b'*\x86H\x86\xf7cd\x01\x07'

    CSSM_APPLE_TP_OCSP_OPTS_VERSION = 0
    CSSM_TP_ACTION_OCSP_DISABLE_NET = 0x00000004
    CSSM_TP_ACTION_OCSP_CACHE_READ_DISABLE = 0x00000008

    CSSM_APPLE_TP_CRL_OPTS_VERSION = 0

    errSecVerifyFailed = -67808
    errSecNoTrustSettings = -25263
    errSecItemNotFound = -25300
    errSecInvalidTrustSettings = -25262

    kSecPaddingNone = 0
    kSecPaddingPKCS1 = 1

    CSSM_KEYUSE_SIGN = 0x00000004
    CSSM_KEYUSE_VERIFY = 0x00000008

    CSSM_ALGID_DH = 2
    CSSM_ALGID_RSA = 42
    CSSM_ALGID_DSA = 43
    CSSM_ALGID_ECDSA = 73
    CSSM_KEYATTR_PERMANENT = 0x00000001
    CSSM_KEYATTR_EXTRACTABLE = 0x00000020



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import sys

from ._errors import pretty_message
from ._types import type_name, byte_cls

if sys.platform == 'darwin':
    from ._mac.util import rand_bytes
elif sys.platform == 'win32':
    from ._win.util import rand_bytes
else:
    from ._openssl.util import rand_bytes


__all__ = [
    'constant_compare',
    'rand_bytes',
]


def constant_compare(a, b):
    """
    Compares two byte strings in constant time to see if they are equal

    :param a:
        The first byte string

    :param b:
        The second byte string

    :return:
        A boolean if the two byte strings are equal
    """

    if not isinstance(a, byte_cls):
        raise TypeError(pretty_message(
            '''
            a must be a byte string, not %s
            ''',
            type_name(a)
        ))

    if not isinstance(b, byte_cls):
        raise TypeError(pretty_message(
            '''
            b must be a byte string, not %s
            ''',
            type_name(b)
        ))

    if len(a) != len(b):
        return False

    if sys.version_info < (3,):
        a = [ord(char) for char in a]
        b = [ord(char) for char in b]

    result = 0
    for x, y in zip(a, b):
        result |= x ^ y
    return result == 0



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

from .._ffi import (
    buffer_from_bytes,
    byte_string_from_buffer,
    deref,
    is_null,
    new,
    register_ffi,
)

from cffi import FFI


__all__ = [
    'CFHelpers',
    'CoreFoundation',
]


ffi = FFI()
ffi.cdef("""
    typedef bool Boolean;
    typedef long CFIndex;
    typedef unsigned long CFStringEncoding;
    typedef unsigned long CFNumberType;
    typedef unsigned long CFTypeID;

    typedef void *CFTypeRef;
    typedef CFTypeRef CFArrayRef;
    typedef CFTypeRef CFDataRef;
    typedef CFTypeRef CFStringRef;
    typedef CFTypeRef CFNumberRef;
    typedef CFTypeRef CFBooleanRef;
    typedef CFTypeRef CFDictionaryRef;
    typedef CFTypeRef CFErrorRef;
    typedef CFTypeRef CFAllocatorRef;

    typedef struct {
        CFIndex version;
        void *retain;
        void *release;
        void *copyDescription;
        void *equal;
        void *hash;
    } CFDictionaryKeyCallBacks;

    typedef struct {
        CFIndex version;
        void *retain;
        void *release;
        void *copyDescription;
        void *equal;
    } CFDictionaryValueCallBacks;

    typedef struct {
        CFIndex version;
        void *retain;
        void *release;
        void *copyDescription;
        void *equal;
    } CFArrayCallBacks;

    CFIndex CFDataGetLength(CFDataRef theData);
    const char *CFDataGetBytePtr(CFDataRef theData);
    CFDataRef CFDataCreate(CFAllocatorRef allocator, const char *bytes, CFIndex length);

    CFDictionaryRef CFDictionaryCreate(CFAllocatorRef allocator, const void **keys, const void **values,
                    CFIndex numValues, const CFDictionaryKeyCallBacks *keyCallBacks,
                    const CFDictionaryValueCallBacks *valueCallBacks);
    CFIndex CFDictionaryGetCount(CFDictionaryRef theDict);

    const char *CFStringGetCStringPtr(CFStringRef theString, CFStringEncoding encoding);
    Boolean CFStringGetCString(CFStringRef theString, char *buffer, CFIndex bufferSize, CFStringEncoding encoding);
    CFStringRef CFStringCreateWithCString(CFAllocatorRef alloc, const char *cStr, CFStringEncoding encoding);

    CFNumberRef CFNumberCreate(CFAllocatorRef allocator, CFNumberType theType, const void *valuePtr);

    CFStringRef CFCopyTypeIDDescription(CFTypeID type_id);

    void CFRelease(CFTypeRef cf);
    void CFRetain(CFTypeRef cf);

    CFStringRef CFErrorCopyDescription(CFErrorRef err);
    CFStringRef CFErrorGetDomain(CFErrorRef err);
    CFIndex CFErrorGetCode(CFErrorRef err);

    Boolean CFBooleanGetValue(CFBooleanRef boolean);

    CFTypeID CFDictionaryGetTypeID(void);
    CFTypeID CFNumberGetTypeID(void);
    CFTypeID CFStringGetTypeID(void);
    CFTypeID CFDataGetTypeID(void);

    CFArrayRef CFArrayCreate(CFAllocatorRef allocator, const void **values, CFIndex numValues,
                    const CFArrayCallBacks *callBacks);
    CFIndex CFArrayGetCount(CFArrayRef theArray);
    CFTypeRef CFArrayGetValueAtIndex(CFArrayRef theArray, CFIndex idx);
    CFNumberType CFNumberGetType(CFNumberRef number);
    Boolean CFNumberGetValue(CFNumberRef number, CFNumberType theType, void *valuePtr);
    CFIndex CFDictionaryGetKeysAndValues(CFDictionaryRef theDict, const void **keys, const void **values);
    CFTypeID CFGetTypeID(CFTypeRef cf);

    extern CFAllocatorRef kCFAllocatorDefault;
    extern CFArrayCallBacks kCFTypeArrayCallBacks;
    extern CFBooleanRef kCFBooleanTrue;
    extern CFDictionaryKeyCallBacks kCFTypeDictionaryKeyCallBacks;
    extern CFDictionaryValueCallBacks kCFTypeDictionaryValueCallBacks;
""")

core_foundation_path = '/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation'

CoreFoundation = ffi.dlopen(core_foundation_path)
register_ffi(CoreFoundation, ffi)

kCFNumberCFIndexType = 14
kCFStringEncodingUTF8 = 0x08000100


class CFHelpers():
    """
    Namespace for core foundation helpers
    """

    _native_map = {}

    @classmethod
    def register_native_mapping(cls, type_id, callback):
        """
        Register a function to convert a core foundation data type into its
        equivalent in python

        :param type_id:
            The CFTypeId for the type

        :param callback:
            A callback to pass the CFType object to
        """

        cls._native_map[int(type_id)] = callback

    @staticmethod
    def cf_number_to_number(value):
        """
        Converts a CFNumber object to a python float or integer

        :param value:
            The CFNumber object

        :return:
            A python number (float or integer)
        """

        type_ = CoreFoundation.CFNumberGetType(value)
        type_name_ = {
            1: 'int8_t',      # kCFNumberSInt8Type
            2: 'in16_t',      # kCFNumberSInt16Type
            3: 'int32_t',     # kCFNumberSInt32Type
            4: 'int64_t',     # kCFNumberSInt64Type
            5: 'float',       # kCFNumberFloat32Type
            6: 'double',      # kCFNumberFloat64Type
            7: 'char',        # kCFNumberCharType
            8: 'short',       # kCFNumberShortType
            9: 'int',         # kCFNumberIntType
            10: 'long',       # kCFNumberLongType
            11: 'long long',  # kCFNumberLongLongType
            12: 'float',      # kCFNumberFloatType
            13: 'double',     # kCFNumberDoubleType
            14: 'long',       # kCFNumberCFIndexType
            15: 'int',        # kCFNumberNSIntegerType
            16: 'double',     # kCFNumberCGFloatType
        }[type_]
        output = new(CoreFoundation, type_name_ + ' *')
        CoreFoundation.CFNumberGetValue(value, type_, output)
        return deref(output)

    @staticmethod
    def cf_dictionary_to_dict(dictionary):
        """
        Converts a CFDictionary object into a python dictionary

        :param dictionary:
            The CFDictionary to convert

        :return:
            A python dict
        """

        dict_length = CoreFoundation.CFDictionaryGetCount(dictionary)

        keys = new(CoreFoundation, 'CFTypeRef[%s]' % dict_length)
        values = new(CoreFoundation, 'CFTypeRef[%s]' % dict_length)
        CoreFoundation.CFDictionaryGetKeysAndValues(
            dictionary,
            keys,
            values
        )

        output = {}
        for index in range(0, dict_length):
            output[CFHelpers.native(keys[index])] = CFHelpers.native(values[index])

        return output

    @classmethod
    def native(cls, value):
        """
        Converts a CF* object into its python equivalent

        :param value:
            The CF* object to convert

        :return:
            The native python object
        """

        type_id = CoreFoundation.CFGetTypeID(value)
        if type_id in cls._native_map:
            return cls._native_map[type_id](value)
        else:
            return value

    @staticmethod
    def cf_string_to_unicode(value):
        """
        Creates a python unicode string from a CFString object

        :param value:
            The CFString to convert

        :return:
            A python unicode string
        """

        string_ptr = CoreFoundation.CFStringGetCStringPtr(
            value,
            kCFStringEncodingUTF8
        )
        string = None if is_null(string_ptr) else ffi.string(string_ptr)
        if string is None:
            buffer = buffer_from_bytes(1024)
            result = CoreFoundation.CFStringGetCString(
                value,
                buffer,
                1024,
                kCFStringEncodingUTF8
            )
            if not result:
                raise OSError('Error copying C string from CFStringRef')
            string = byte_string_from_buffer(buffer)
        if string is not None:
            string = string.decode('utf-8')
        return string

    @staticmethod
    def cf_string_from_unicode(string):
        """
        Creates a CFStringRef object from a unicode string

        :param string:
            The unicode string to create the CFString object from

        :return:
            A CFStringRef
        """

        return CoreFoundation.CFStringCreateWithCString(
            CoreFoundation.kCFAllocatorDefault,
            string.encode('utf-8'),
            kCFStringEncodingUTF8
        )

    @staticmethod
    def cf_data_to_bytes(value):
        """
        Extracts a bytestring from a CFData object

        :param value:
            A CFData object

        :return:
            A byte string
        """

        start = CoreFoundation.CFDataGetBytePtr(value)
        num_bytes = CoreFoundation.CFDataGetLength(value)
        return ffi.buffer(start, num_bytes)[:]

    @staticmethod
    def cf_data_from_bytes(bytes_):
        """
        Creates a CFDataRef object from a byte string

        :param bytes_:
            The data to create the CFData object from

        :return:
            A CFDataRef
        """

        return CoreFoundation.CFDataCreate(
            CoreFoundation.kCFAllocatorDefault,
            bytes_,
            len(bytes_)
        )

    @staticmethod
    def cf_dictionary_from_pairs(pairs):
        """
        Creates a CFDictionaryRef object from a list of 2-element tuples
        representing the key and value. Each key should be a CFStringRef and each
        value some sort of CF* type.

        :param pairs:
            A list of 2-element tuples

        :return:
            A CFDictionaryRef
        """

        length = len(pairs)
        keys = []
        values = []
        for pair in pairs:
            key, value = pair
            keys.append(key)
            values.append(value)
        return CoreFoundation.CFDictionaryCreate(
            CoreFoundation.kCFAllocatorDefault,
            keys,
            values,
            length,
            ffi.addressof(CoreFoundation.kCFTypeDictionaryKeyCallBacks),
            ffi.addressof(CoreFoundation.kCFTypeDictionaryValueCallBacks)
        )

    @staticmethod
    def cf_array_from_list(values):
        """
        Creates a CFArrayRef object from a list of CF* type objects.

        :param values:
            A list of CF* type object

        :return:
            A CFArrayRef
        """

        length = len(values)
        return CoreFoundation.CFArrayCreate(
            CoreFoundation.kCFAllocatorDefault,
            values,
            length,
            ffi.addressof(CoreFoundation.kCFTypeArrayCallBacks)
        )

    @staticmethod
    def cf_number_from_integer(integer):
        """
        Creates a CFNumber object from an integer

        :param integer:
            The integer to create the CFNumber for

        :return:
            A CFNumber
        """

        integer_as_long = ffi.new('long *', integer)
        return CoreFoundation.CFNumberCreate(
            CoreFoundation.kCFAllocatorDefault,
            kCFNumberCFIndexType,
            integer_as_long
        )
